///////////////////////////////////////////////////////////////////////////////
//
//	(C) Copyright 1995 - 2001 OSR Open Systems Resources, Inc.
//  Copyright (c) 1990-2000    Microsoft Corporation All Rights Reserved
//	All Rights Reserved
//
//	This software is part of a licensed software product and may
//	only be used or copied in accordance with the terms of that license.
//
//	OSR Open Systems Resources, Inc.
//	105 Route 101A Suite 19
//	Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//	email bugs to: bugs@osr.com
//
//
//	MODULE:
//
//		$Workfile: nothingpower.c $
//
//	ABSTRACT:
//
//      This module handles Power management calls for the FDO.   We're not 
//      really doing anything in here, just null power management.    It works.....
//
//	AUTHOR:
//
//		Open Systems Resources, Inc.
// 
//	REVISION:   
//
//		$Revision: 1 $
//
///////////////////////////////////////////////////////////////////////////////
#include <ntddk.h>
#include "nothing.h"
#include "stdio.h"
#include <stddef.h>

VOID nothingFDOPowerTransitionPoRequestComplete (IN PDEVICE_OBJECT DeviceObject,
                                             IN UCHAR MinorFunction,
                                             IN POWER_STATE DevicePowerState,
                                             IN PVOID Irp, 
                                             IN PIO_STATUS_BLOCK IoStatus);
NTSTATUS nothingFDOPowerComplete(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp,
                           IN PVOID Context);

PUCHAR PrintSystemPowerState(SYSTEM_POWER_STATE State)
{
    PUCHAR string;

    switch(State) {

        case PowerSystemUnspecified: // = 0,
            string = (PUCHAR) "PowerSystemUnspecified";
            break;

        case PowerSystemWorking: //,
            string = (PUCHAR) "PowerSystemWorking";
            break;

        case PowerSystemSleeping1: //,
            string = (PUCHAR) "PowerSystemSleeping1";
            break;

        case PowerSystemSleeping2: //,
            string = (PUCHAR) "PowerSystemSleeping2";
            break;

        case PowerSystemSleeping3: //,
            string = (PUCHAR) "PowerSystemSleeping3";
            break;

        case PowerSystemHibernate: //,
            string = (PUCHAR) "PowerSystemHibernate";
            break;

        case PowerSystemShutdown: //,
            string = (PUCHAR) "PowerSystemShutdown";
            break;

        case PowerSystemMaximum: //
            string = (PUCHAR) "PowerSystemMaximum";
            break;

        default:
            string = (PUCHAR) "Unknown Power State ?????? Error";
            break;


    }

    return string;
}

PUCHAR PrintPowerAction(POWER_ACTION Action)
{
    PUCHAR string;

    switch(Action) {

        case PowerActionNone: // = 0,
            string = (PUCHAR) "PowerActionNone";
            break;

        case PowerActionReserved: //,
            string = (PUCHAR) "PowerActionReserved";
            break;

        case PowerActionSleep: //,
            string = (PUCHAR) "PowerActionSleep";
            break;

        case PowerActionHibernate: //,
            string = (PUCHAR) "PowerActionHibernate";
            break;

        case PowerActionShutdown: //,
            string = (PUCHAR) "PowerActionShutdown";
            break;

        case PowerActionShutdownReset: //,
            string = (PUCHAR) "PowerActionShutdownReset";
            break;

        case PowerActionShutdownOff: //,
            string = (PUCHAR) "PowerActionShutdownOff";
            break;

        case PowerActionWarmEject: //
            string = (PUCHAR) "PowerActionWarmEject";
            break;

        default:
            string = (PUCHAR) "Unknown Power Action ???? Error";
            break;

    }

    return string;
}

PUCHAR PrintDevicePowerState(DEVICE_POWER_STATE State)
{
    PUCHAR string;

    switch(State) {

        case PowerDeviceUnspecified: // = 0,
            string = (PUCHAR) "PowerDeviceUnspecified";
            break;

        case PowerDeviceD0: //,
            string = (PUCHAR) "PowerDeviceD0";
            break;

        case PowerDeviceD1: //,
            string = (PUCHAR) "PowerDeviceD1";
            break;

        case PowerDeviceD2: //,
            string = (PUCHAR) "PowerDeviceD2";
            break;

        case PowerDeviceD3: //,
            string = (PUCHAR) "PowerDeviceD3";
            break;

        case PowerDeviceMaximum: //
            string = (PUCHAR) "PowerDeviceMaximum";
            break;

        default:
            string = (PUCHAR) "Unknown Device Power State ???? Error";
            break;

    }

    return string;
}

PUCHAR PrintPowerStateType(POWER_STATE_TYPE Type)
{
    PUCHAR string;

    switch(Type) {
        case SystemPowerState:// = 0,
            string = (PUCHAR) "SystemPowerState";
            break;

        case DevicePowerState: //
            string = (PUCHAR) "DevicePowerState";
            break;

        default:
            string = (PUCHAR) "Unknown POWER_STATE_TYPE";
            break;

    }

    return string;
}


///////////////////////////////////////////////////////////////////////////////
//
//  nothingFDOPower
//
//      Handles power Irps sent to the FDO.
//      This driver is power policy owner for
//      the bus driver itself (not the devices on the bus).
//
//  INPUTS:
//
//      DeviceObject - Pointer to the device object.
//      Irp          - Pointer to the irp.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      STATUS_SUCCESS if okay, an error otherwise.
//
//  IRQL:
//
//      This routine is called at any IRQL PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS nothingFDOPower (PNOTHING_DEVICE_EXT FdoData,PIRP Irp)
{
    NTSTATUS            status;
    POWER_STATE         powerState;
    POWER_STATE_TYPE    powerType;
    PIO_STACK_LOCATION  stack;
    PLIST_ENTRY         entry;
    BOOLEAN             needCallback = FALSE;
    BOOLEAN             wait = FALSE;
    
    DbgPrint("nothingFDOPower Entered.\n");

    stack = IoGetCurrentIrpStackLocation (Irp);
    powerType = stack->Parameters.Power.Type;
    powerState = stack->Parameters.Power.State;

    //
    // What do we have.....
    //

    switch (stack->MinorFunction) {

        case IRP_MN_SET_POWER:
            DbgPrint("Setting %s state to %s\n",
                         ((powerType == SystemPowerState) ?  "System" : "Device"),
                         ((powerType == SystemPowerState) ? PrintSystemPowerState(powerState.SystemState) :
                                                            PrintDevicePowerState(powerState.DeviceState)));

            switch (powerType) {

                case DevicePowerState:

                    status = Irp->IoStatus.Status = STATUS_SUCCESS;

                    if ( PowerDeviceD0 < powerState.DeviceState) {

                        //
                        // Powering down
                        //

                        PoSetPowerState (FdoData->DeviceObject, powerType, powerState);
                        FdoData->DevicePowerState = powerState.DeviceState;

                    } else {
                        //
                        // Powering Up
                        //
                        needCallback = TRUE;
                    }

                    break;

                case SystemPowerState:

                    status = Irp->IoStatus.Status = STATUS_SUCCESS;

                    if (PowerSystemWorking < powerState.SystemState) {
                        //
                        // Powering down
                        //
                        // Since our S-state to D-state mapping is D3 for all the S states
                        // except S0. We always request a D3 irp in response to this S irp.  
                        // The D3 irp must be completed before sending this S irp down the 
                        // stack. We will send the S irp down the stack when
                        // OsrFDOPowerTransitionPoRequestComplete is called.
                        //

                        switch(powerState.SystemState) {

                            case PowerSystemShutdown:
                                DbgPrint("FDO PowerDown - Notifying User Code\n");
                                break;

                            default:
                                break;

                        }

                        //
                        // We don't need to increment our IO count b/c we incremented it
                        // at the beginning of this function and won't decrement it until
                        // the S Irp completes
                        // 
                        IoMarkIrpPending (Irp);
                
                        powerState.DeviceState = PowerDeviceD3;

                        //
                        // PoRequestPowerIrp return STATUS_PENDING
                        // when the IRP is successfully sent.
                        //
                
                        status = PoRequestPowerIrp (FdoData->DeviceObject,
                                           IRP_MN_SET_POWER,
                                           powerState,
                                           nothingFDOPowerTransitionPoRequestComplete,
                                           Irp,
                                           NULL);  // no IRP

                                   
                        return status;

                    } else {
                        //
                        // Powering Up
                        //
                        needCallback = TRUE;
                    }
                    break;
            }

            break;

        case IRP_MN_WAIT_WAKE:
        case IRP_MN_QUERY_POWER:
            break;

        default:
            break;
    }

    IoCopyCurrentIrpStackLocationToNext (Irp);

    //
    // See if we set the needCallback flag.  If we did, we need to know
    // when this Power Irp completes.   
    //

    if (needCallback) {

        //
        // We need completion notification.  Make sure that whatever
        // we did before this was successful.
        //

//        OSRASSERT (STATUS_SUCCESS == status);

        //
        // Set up the completion routine.
        //

        IoSetCompletionRoutine (Irp,nothingFDOPowerComplete,NULL,TRUE,TRUE,TRUE);

        //
        // Pass the Power Irp on....
        //

        status = PoCallDriver (FdoData->DeviceToSendIrpsTo, Irp);

    } else {

        //
        // We're not interested in this irp, so we'll start
        // the next one and pass this one down.
        //
        
        PoStartNextPowerIrp (Irp);

        status =  PoCallDriver (FdoData->DeviceToSendIrpsTo, Irp);
        
    }

    return status;
}


///////////////////////////////////////////////////////////////////////////////
//
//  nothingFDOPowerComplete
//
//      Handles power Irps completions
//
//  INPUTS:
//
//      DeviceObject - Pointer to the device object.
//      Irp          - Pointer to the irp.
//      Context      - Completion context
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      STATUS_MORE_PROCESSING_REQUIRED if IRP to be retained, some
//      other status otherwise.
//
//  IRQL:
//
//    This routine is called at any IRQL <= DISPATCH_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS nothingFDOPowerComplete(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp,
                           IN PVOID Context)
{
    POWER_STATE         powerState;
    POWER_STATE_TYPE    powerType;
    PIO_STACK_LOCATION  stack;
    PNOTHING_DEVICE_EXT   fdoData;
    NTSTATUS            status = STATUS_SUCCESS;
    
    UNREFERENCED_PARAMETER (Context);

    //
    // Get the information that we need out of the irp.
    //

    fdoData = (PNOTHING_DEVICE_EXT) DeviceObject->DeviceExtension;
    stack = IoGetCurrentIrpStackLocation (Irp);
    powerType = stack->Parameters.Power.Type;
    powerState = stack->Parameters.Power.State;
    
    DbgPrint("Entered nothingFDOPowerComplete\n");

    //
    // what did we find completed....
    //

    switch (stack->MinorFunction) {

        case IRP_MN_SET_POWER:

            switch (powerType) {
        
                case DevicePowerState:

                    //
                    // Device Power State change....
                    //

                    fdoData->DevicePowerState = powerState.DeviceState;

                    PoSetPowerState (fdoData->DeviceObject, powerType, powerState);
                    break;

                case SystemPowerState:
            
                    //
                    // System Power State change....
                    //

                    if (PowerSystemWorking > fdoData->SystemPowerState) {

                        //
                        // We're Powering Down...
                        //
                        // We are on the completion end of an S irp.  (The D3 power irp
                        // has already been sent and completed down this stack.)  The
                        // remaining thing to do is set the state in the extension, then
                        // decrement the IoCount that was incremented when we first got
                        // the irp (this is done at the end of this function).
                        //

                        DbgPrint("powering down\n");

                        fdoData->SystemPowerState = powerState.SystemState;

                        PoSetPowerState (fdoData->DeviceObject,
                                         stack->Parameters.Power.Type,
                                         stack->Parameters.Power.State);

                    } else {

                        //
                        // We're Powering Up...
                        //
                        // Request a D power irp for ourself. Do not complete this S irp
                        // until the D irp has been completed.(Completion of the S irp
                        // is done in OsrFDOPowerTransitionPoRequestComplete). 
                        // Decrementing the IO count will happen in the same function.
                        //
    
                        DbgPrint("powering up\n");

                        powerState.DeviceState = PowerDeviceD0;
                        PoRequestPowerIrp (fdoData->DeviceObject,
                                           IRP_MN_SET_POWER,
                                           powerState,
                                           nothingFDOPowerTransitionPoRequestComplete,
                                           Irp, 
                                           NULL); // no return Irp
    
                        //
                        // Inform the IO subsystem not to touch the irp here 
                        //
                        // The irp might completed by the time we get here, so call
                        // PoStartNextPowerIrp in the PO irp completion function.
                        //

                        status = STATUS_MORE_PROCESSING_REQUIRED; 
                    }
                    break;
            }
            break;

        case IRP_MN_QUERY_POWER:
            break;

        default:
            break;
    }

    if(NT_SUCCESS(status)) {

        PoStartNextPowerIrp (Irp);
    }
    return status;
}


///////////////////////////////////////////////////////////////////////////////
//
//  nothingFDOPowerTransitionPoRequestComplete
//
//      This completion routine is set to device set-power irp
//      generated by the driver. Its purpose is to note the
//      system state and complete the SystemPower IRP during
//      power up or send the SystemPower IRP down during power down.
//
//  INPUTS:
//
//      DeviceObject        - Pointer to the device object.
//      MinorFunction       - Power minor function
//      DevicePowerState    - current device power state
//      Irp                 - Pointer to the irp.
//      IoStatus            - Status block
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  IRQL:
//
//    This routine is called at any IRQL <= DISPATCH_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
VOID nothingFDOPowerTransitionPoRequestComplete (IN PDEVICE_OBJECT DeviceObject,
                                             IN UCHAR MinorFunction,
                                             IN POWER_STATE DevicePowerState,
                                             IN PVOID Irp, 
                                             IN PIO_STATUS_BLOCK IoStatus)
{
    PIO_STACK_LOCATION  stack;
    PNOTHING_DEVICE_EXT   fdoData;
    PIRP                SystemStateIrp = (PIRP) Irp;

    UNREFERENCED_PARAMETER (MinorFunction);
    UNREFERENCED_PARAMETER (IoStatus);


    fdoData = (PNOTHING_DEVICE_EXT) DeviceObject->DeviceExtension;
    stack = IoGetCurrentIrpStackLocation (SystemStateIrp);
    
    DbgPrint("Entered nothingFDOPowerTransitionPoRequestComplete\n");

    if (DevicePowerState.DeviceState == PowerDeviceD0) {

        NTSTATUS status = STATUS_SUCCESS;

        //
        // We are powering up (the D0 Irp just completed).  Since we sent the
        // S irp down the stack and requested the D irp on the way back up the
        // stack, just complete the S irp now
        //

        PoSetPowerState (DeviceObject,
                         stack->Parameters.Power.Type,
                         stack->Parameters.Power.State);
    
        fdoData->SystemPowerState = stack->Parameters.Power.State.SystemState;

        SystemStateIrp->IoStatus.Status = status;
        PoStartNextPowerIrp (SystemStateIrp);
        SystemStateIrp->IoStatus.Status = status;
        IoCompleteRequest (SystemStateIrp, IO_NO_INCREMENT);

        //
        // From OsrFDOPower when we originally received the IRP
        //

    } else {
        //
        // We are powering down (the D3 Irp just completed).  Since we requested
        // the D irp before sending the S irp down the stack, we must send it 
        // down now.  We will catch the S irp on the way back up to record the 
        // S state
        //
        //OSRASSERT (DevicePowerState.DeviceState == PowerDeviceD3);
    
        IoCopyCurrentIrpStackLocationToNext (SystemStateIrp);

        IoSetCompletionRoutine (SystemStateIrp,
                                nothingFDOPowerComplete,
                                NULL, TRUE, TRUE, TRUE);
    
        PoCallDriver (fdoData->DeviceToSendIrpsTo, SystemStateIrp);
    }
}


///////////////////////////////////////////////////////////////////////////////
//
//  nothingPower
//
//      Handles power Irps sent to both FDO and child PDOs.
//
//  INPUTS:
//
//      DeviceObject - Pointer to the device object.
//
//      Irp          - Pointer to the irp.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      STATUS_SUCCESS if okay, an error otherwise.
//
//  IRQL:
//
//    This routine is called at any IRQL PASSIVE_LEVEL.
//
//  NOTES:
//
///////////////////////////////////////////////////////////////////////////////
NTSTATUS NothingPower(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp)
{
    PIO_STACK_LOCATION  irpStack;
    NTSTATUS            status;
    PNOTHING_DEVICE_EXT   fdoData;

    DbgPrint("nothingPower Entered.\n");
    

    status = STATUS_SUCCESS;
    irpStack = IoGetCurrentIrpStackLocation (Irp);

    fdoData = (PNOTHING_DEVICE_EXT) DeviceObject->DeviceExtension;

    status = nothingFDOPower (fdoData,Irp);

    return status;
}


