///////////////////////////////////////////////////////////////////////////////
//
//    (C) Copyright 1995 - 2001 OSR Open Systems Resources, Inc.
//    (C) Copyright 2000 MicroSoft, Inc.
//    All Rights Reserved
//
//    This sofware is supplied for instructional purposes only.
//
//    OSR Open Systems Resources, Inc. (OSR) expressly disclaims any warranty
//    for this software.  THIS SOFTWARE IS PROVIDED  "AS IS" WITHOUT WARRANTY
//    OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION,
//    THE IMPLIED WARRANTIES OF MECHANTABILITY OR FITNESS FOR A PARTICULAR
//    PURPOSE.  THE ENTIRE RISK ARISING FROM THE USE OF THIS SOFTWARE REMAINS
//    WITH YOU.  OSR's entire liability and your exclusive remedy shall not
//    exceed the price paid for this material.  In no event shall OSR or its
//    suppliers be liable for any damages whatsoever (including, without
//    limitation, damages for loss of business profit, business interruption,
//    loss of business information, or any other pecuniary loss) arising out
//    of the use or inability to use this software, even if OSR has been
//    advised of the possibility of such damages.  Because some states/
//    jurisdictions do not allow the exclusion or limitation of liability for
//    consequential or incidental damages, the above limitation may not apply
//    to you.
//
//    OSR Open Systems Resources, Inc.
//    105 Route 101A Suite 19
//    Amherst, NH 03031  (603) 595-6500 FAX: (603) 595-6503
//    email bugs to: bugs@osr.com
//
//
//    MODULE:
//
//        proppage.cpp -- Nothing driver Property Page
//
//    ABSTRACT:
//
//      This file contains the code for property page that talks to the Nothing
//      Driver.
//
//    AUTHOR(S):
//
//        OSR Open Systems Resources, Inc.
// 
//    REVISION:   
//
///////////////////////////////////////////////////////////////////////////////

//#include "stdafx.h"
#include <windows.h>
//#include <devioctl.h>
#include <setupapi.h>
#include "proppage.h"
#include "prvprop.h"
#include "resource.h"
#include <winioctl.h>
#include <nothing_ioctl.h>

#if DBG
#define NothingOutputDebugString(a) OutputDebugString(a)
#else
#define NothingOutputDebugString(a)
#endif

HINSTANCE   ghInstance = NULL;

typedef struct _NOTHING_PROP_INTF
{

    NOTHING_FEATURES                    NothingFeatures;
    PSP_DEVICE_INTERFACE_DETAIL_DATA    PIntfDetail;

} NOTHING_PROP_INTF, *PNOTHING_PROP_INTF;


#ifdef _DEBUG

///////////////////////////////////////////////////////////////////////////////
//
//  DbgError
//
//      This function prints an error message.
//      It prints first the string passed and then the error that it gets with
//      GetLastError as a string.
//
//  INPUTS:
//
//      szMsg - message to print
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
void DbgError (LPSTR szMsg)
{
    static TCHAR szErrorMessage[1024];

    // Get the error message from the system.
    FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
				   NULL,
				   GetLastError (),
				   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				   szErrorMessage,
				   sizeof (szErrorMessage),
                   NULL);

	// Print the msg + error + \n\r.
    NothingOutputDebugString (szMsg);
    NothingOutputDebugString (szErrorMessage);
    NothingOutputDebugString ("\n\r");

    // This is for the guys who don't have a debugger.
    MessageBox (NULL, szErrorMessage, szMsg, MB_OK | MB_ICONSTOP);
}
#else
#define DbgError(a) 0;
#endif


///////////////////////////////////////////////////////////////////////////////
//
//  DllMain
//
//      This function is called when processes or threads attach or detach
//      from the DLL.
//
//  INPUTS:
//
//      hModule - handle of the module attaching.
//      ul_reason_for_call - why DLL is being called.
//      lpReserved - reserved.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      TRUE, if operation successful.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    switch (ul_reason_for_call) {

		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
    }

    ghInstance = (HINSTANCE) hModule;

    NothingOutputDebugString("DllMain Entered\n");

    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  GetDeviceInterfaceDetail
//
//      This function gets called by the property page provider (in this module) to
//      get the device interface details. The device interface detail contains a
//      path to the device driver that can be used to open the device driver.
//      When we parse the driver we look for the interface the Nothing WDM driver
//      exposes.
//
//  INPUTS:
//
//      pPropPageRequest           - points to SP_PROPSHEETPAGE_REQUEST
//      pDeviceInterfaceDetailData - device interface details returned.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      TRUE, if operation successful, FALSE otherwise.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL GetDeviceInterfaceDetail (PSP_PROPSHEETPAGE_REQUEST pPropPageRequest,
                               PSP_DEVICE_INTERFACE_DETAIL_DATA *ppDeviceInterfaceDetailData)
{
    BOOL                        fSuccess;
    ULONG                       ulDeviceInstanceIdSize = 0;
    PTSTR                       pDeviceInstanceID = NULL;
    HDEVINFO                    hDevInfoWithInterface;
	SP_DEVICE_INTERFACE_DATA    DeviceInterfaceData;
    ULONG                       ulDeviceInterfaceDetailDataSize = 0;

    NothingOutputDebugString("GetDeviceInterfaceDetail Entered\n");

    // Get the device instance id (PnP string).  The first call will retrieve
    // the buffer length in characters.  fSuccess will be FALSE.
    fSuccess = SetupDiGetDeviceInstanceId (pPropPageRequest->DeviceInfoSet,
                                           pPropPageRequest->DeviceInfoData,
                                           NULL,
                                           0,
                                           &ulDeviceInstanceIdSize);
    // Check for error.
    if ((GetLastError () != ERROR_INSUFFICIENT_BUFFER) || (!ulDeviceInstanceIdSize)) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiGetDeviceInstanceId: ");
        return FALSE;
	}

    // Allocate the buffer for the device instance ID (PnP string).
    pDeviceInstanceID = (PTSTR)LocalAlloc (LPTR, ulDeviceInstanceIdSize * sizeof (TCHAR));
    if (!pDeviceInstanceID) {

		DbgError ("GetDeviceInterfaceDetail: LocalAlloc: ");
        return FALSE;
    }
    
    // Now call again, this time with all parameters.
    fSuccess = SetupDiGetDeviceInstanceId (pPropPageRequest->DeviceInfoSet,
                                           pPropPageRequest->DeviceInfoData,
                                           pDeviceInstanceID,
                                           ulDeviceInstanceIdSize,
                                           NULL);
    // Check for error.
    if (!fSuccess) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiGetDeviceInstanceId: ");
        LocalFree (pDeviceInstanceID);
        return FALSE;
	}

    // Now we can get the handle to the dev info with interface.
    // We parse the device specifically for the NOTHINGID_PRIVATE interface.
    hDevInfoWithInterface = SetupDiGetClassDevs ((struct _GUID*) &NOTHINGID_PRIVATE,
                                                 pDeviceInstanceID,
                                                 NULL,
                                                 DIGCF_DEVICEINTERFACE);
    // We don't need pDeviceInstanceID anymore.
    LocalFree (pDeviceInstanceID);

    // Check for error.
    if (hDevInfoWithInterface == INVALID_HANDLE_VALUE) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiGetClassDevs: ");
        return FALSE;
    }

    // Go through the list of device interfaces of this device.
    // We assume that there is only one device interface and
    // we will store the device details in our private structure.
	DeviceInterfaceData.cbSize = sizeof (DeviceInterfaceData);
    fSuccess = SetupDiEnumDeviceInterfaces (hDevInfoWithInterface,
					    				    NULL,
											(struct _GUID*) &NOTHINGID_PRIVATE,
											0,
											&DeviceInterfaceData);
    // Check for error.
    if (!fSuccess) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiEnumDeviceInterfaces: ");
        SetupDiDestroyDeviceInfoList (hDevInfoWithInterface);
        return FALSE;
	}

    // Get the details for this device interface.  The first call will retrieve
    // the buffer length in characters.  fSuccess will be FALSE.
    fSuccess = SetupDiGetDeviceInterfaceDetail (hDevInfoWithInterface,
                                                &DeviceInterfaceData,
                                                NULL,
                                                0,
                                                &ulDeviceInterfaceDetailDataSize,
                                                NULL);
    // Check for error.
    if ((GetLastError () != ERROR_INSUFFICIENT_BUFFER) || (!ulDeviceInterfaceDetailDataSize)) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiGetDeviceInterfaceDetail: ");
        SetupDiDestroyDeviceInfoList (hDevInfoWithInterface);
        return FALSE;
	}

    // Allocate the buffer for the device interface detail data.
    if (!(*ppDeviceInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)
        LocalAlloc (LPTR, ulDeviceInterfaceDetailDataSize))) {

		DbgError ("GetDeviceInterfaceDetail: LocalAlloc: ");
        SetupDiDestroyDeviceInfoList (hDevInfoWithInterface);
        return FALSE;
    }
    // The size contails only the structure, not the additional path.
    (*ppDeviceInterfaceDetailData)->cbSize = sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);

    // Get the details for this device interface, this time with all paramters.
    fSuccess = SetupDiGetDeviceInterfaceDetail (hDevInfoWithInterface,
                                                &DeviceInterfaceData,
                                                *ppDeviceInterfaceDetailData,
                                                ulDeviceInterfaceDetailDataSize,
                                                NULL,
                                                NULL);
    // We don't need the handle anymore.
    SetupDiDestroyDeviceInfoList (hDevInfoWithInterface);

    if (!fSuccess) {

		DbgError ("GetDeviceInterfaceDetail: SetupDiGetDeviceInterfaceDetail: ");
        LocalFree (*ppDeviceInterfaceDetailData), *ppDeviceInterfaceDetailData = NULL;
        return FALSE;
	}

    NothingOutputDebugString("GetDeviceInterfaceDetail Exit\n");

    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  SetDlgControls
//
//      This function gets called when the property sheet page gets created by
//      "NothingPropPage_OnInitDialog". It initializes the different dialog controls that
//      get displayed.
//
//  INPUTS:
//
//      hWnd           - handle to the dialog window
//      PNothingFeatures  - pointer to NOTHING Features structure
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      None.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
void SetDlgControls (HWND hWnd, PNOTHING_FEATURES PNothingFeatures)
{
    UINT checkState = BST_UNCHECKED;

    NothingOutputDebugString("SetDlgControls Enter\n");

    if(PNothingFeatures->DoNothing) {

        NothingOutputDebugString("SetDlgControls Setting DoNothing to Checked\n");
        checkState = BST_CHECKED;

    } else {

        NothingOutputDebugString("SetDlgControls Setting DoNothing to Unchecked\n");

    }

    CheckDlgButton(hWnd, IDC_CHECK_DO_NOTHING, checkState);

    EnableWindow (GetDlgItem (hWnd, IDC_CHECK_DO_NOTHING), TRUE);

    NothingOutputDebugString("SetDlgControls Exit\n");
}


///////////////////////////////////////////////////////////////////////////////
//
//  GetNothingFeatures
//
//      This function gets called by the property page provider (in this module) to
//      get all the Nothing Driver features that are normally not displayed by the drivers.
//      To get the NothingFeatures structure we call the Nothing Driver. As you
//      can see this is fairly easy..
//
//  INPUTS:
//
//      pDeviceInterfaceDetailData - device interface details (path to device driver)
//      PNothingFeatures              - pointer to Nothing features structure.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      TRUE, if successful, FALSE otherwise.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL GetNothingFeatures (PSP_DEVICE_INTERFACE_DETAIL_DATA pDeviceInterfaceDetailData,
                         PNOTHING_PROP_INTF PNothingFeatures)
{
    HANDLE          hDriver;
    ULONG           ulBytesReturned;
    BOOL            fSuccess;

    NothingOutputDebugString("GetNothingFeatures Enter\n");

	// Open the Nothing Driver
    
    hDriver = CreateFile (pDeviceInterfaceDetailData->DevicePath,
                            GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL, OPEN_EXISTING, 0, NULL);
    // Check for error.
    if (hDriver == INVALID_HANDLE_VALUE) {

        DbgError ("GetNothingFeatures: CreateFile: ");
        return FALSE;
    }

    fSuccess = DeviceIoControl (hDriver, IOCTL_OSR_GET_NOTHING_PROPERTY,
                                NULL, 0,
                                &PNothingFeatures->NothingFeatures, sizeof (NOTHING_FEATURES),
                                &ulBytesReturned, NULL);

    // We don't need the handle anymore.
    CloseHandle (hDriver);
    
    // Check for error.
    if (!fSuccess) {

        DbgError ("GetNothingFeatures: DeviceIoControl: ");
        return FALSE;
    }

    NothingOutputDebugString("GetNothingFeatures Exit\n");

    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  NothingPropPage_OnInitDialog
//
//      This function gets called when the property sheet page gets created.  This
//      is the perfect opportunity to initialize the dialog items that get displayed.
//
//  INPUTS:
//
//      ParentHwnd - handle to the dialog window
//      FocusHwnd  - handle to the control that would get the focus.
//      lParam     - initialization parameter (pNothingFeatures).
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      TRUE if focus should be set to FocusHwnd, FALSE if you set the focus yourself
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL NothingPropPage_OnInitDialog (HWND ParentHwnd, HWND FocusHwnd, LPARAM lParam)
{
    PNOTHING_PROP_INTF  pNothingFeatures;
    HCURSOR			    hCursor;

    NothingOutputDebugString("NothingPropPage_OnInitDialog Enter\n");

	// put up the wait cursor
	hCursor = SetCursor (LoadCursor (NULL, IDC_WAIT));

    pNothingFeatures = (PNOTHING_PROP_INTF)((LPPROPSHEETPAGE)lParam)->lParam;

    GetNothingFeatures(pNothingFeatures->PIntfDetail,pNothingFeatures);

    SetDlgControls (ParentHwnd, &pNothingFeatures->NothingFeatures);

    SetWindowLong(ParentHwnd,DWL_USER,(LONG) pNothingFeatures);

    // remove the wait cursor
	SetCursor(hCursor);

    NothingOutputDebugString("NothingPropPage_OnInitDialog Exit\n");

	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  SetNothingFeatures
//
//      This function gets called by the property page provider (in this module) to
//      set all the Nothing Driver features that are normally not displayed by the drivers.
//      To get the NothingFeatures structure we call the Nothing Driver. As you
//      can see this is fairly easy..
//
//  INPUTS:
//
//      pDeviceInterfaceDetailData - device interface details (path to device driver)
//      PNothingFeatures              - pointer to Nothing features structure.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      TRUE, if successful, FALSE otherwise.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL SetNothingFeatures (PSP_DEVICE_INTERFACE_DETAIL_DATA pDeviceInterfaceDetailData,
                         PNOTHING_PROP_INTF PNothingFeatures)
{
    HANDLE          hDriver;
    ULONG           ulBytesReturned;
    BOOL            fSuccess;

    NothingOutputDebugString("SetNothingFeatures Enter\n");

	// Open the Nothing Driver
    
    hDriver = CreateFile (pDeviceInterfaceDetailData->DevicePath,
                            GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL, OPEN_EXISTING, 0, NULL);
    // Check for error.
    if (hDriver == INVALID_HANDLE_VALUE) {

        DbgError ("SetNothingFeatures: CreateFile: ");
        return FALSE;
    }

    fSuccess = DeviceIoControl (hDriver, IOCTL_OSR_SET_NOTHING_PROPERTY,
                                &PNothingFeatures->NothingFeatures, sizeof (NOTHING_FEATURES),
                                NULL, 0,
                                &ulBytesReturned, NULL);

    // We don't need the handle anymore.
    CloseHandle (hDriver);
    
    // Check for error.
    if (!fSuccess) {

        DbgError ("SetNothingFeatures: DeviceIoControl: ");
        return FALSE;
    }

    NothingOutputDebugString("GetNothingFeatures Exit\n");

    return TRUE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  NothingDlgProc
//
//      This callback function gets called by the system whenever something happens
//      to the dialog sheet. Please take a look at the SDK for further information
//      on dialog messages.
//
//  INPUTS:
//
//      hDlg     - handle to the dialog window
//      uMessage - the message
//      wParam   - depending on message sent
//      lParam   - depending on message sent
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      int (depends on message).
//
//  NOTES:
//
//      If it is for our control, the GetWindowLong call will retrieve the 
//      information that we set in the WindowLong.  Namely the pointer to
//      the Nothing Features structure which tells us what driver we're 
//      communicating with.
//
///////////////////////////////////////////////////////////////////////////////
int APIENTRY NothingDlgProc (HWND hDlg, UINT uMessage, WPARAM wParam, LPARAM lParam)
{
    switch (uMessage) {

		// We don't do anything for these messages.
        case WM_COMMAND:
            NothingOutputDebugString("NothingDlgProc WM_COMMAND\n");

            switch(LOWORD(wParam)) {

                case IDC_CHECK_DO_NOTHING: {

                    PNOTHING_PROP_INTF pNothingFeatures = (PNOTHING_PROP_INTF)GetWindowLong(hDlg, DWL_USER);
                    UINT val = IsDlgButtonChecked(hDlg,IDC_CHECK_DO_NOTHING);

                    //
                    // Something has happened to our control.   See if the state of the button
                    // has changed.
                    //

                    switch(val) {

                        case BST_CHECKED:
                            pNothingFeatures->NothingFeatures.DoNothing = TRUE;
                            NothingOutputDebugString("NothingDlgProc Checked\n");
                            break;

                        case BST_UNCHECKED:
                            pNothingFeatures->NothingFeatures.DoNothing = FALSE;
                            NothingOutputDebugString("NothingDlgProc Unchecked\n");
                            break;

                        default:
                            NothingOutputDebugString("NothingDlgProc Indeterminate\n");
                            break;

                    }
                    SetNothingFeatures (pNothingFeatures->PIntfDetail,pNothingFeatures);
                    }
                    break;
                default:
                    break;

            }
            NothingOutputDebugString("NothingDlgProc WM_COMMAND exit\n");
            break;

		case WM_CONTEXTMENU:
		case WM_HELP:
		case WM_NOTIFY:
		case WM_DESTROY:
            break;

        case WM_INITDIALOG:
			return NothingPropPage_OnInitDialog (hDlg, (HWND) wParam, lParam);
	}

	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
//
//  NothingSettingsDlgCallback
//
//      This callback function gets called by the system whenever something happens
//      with the property page. 
//
//  INPUTS:
//
//      hwnd     - handle to the dialog window
//      uMsg     - the message
//      ppsp     - Property Sheet Information
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      int (depends on message).
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
UINT CALLBACK NothingSettingsDlgCallback(HWND hwnd,UINT uMsg,LPPROPSHEETPAGE ppsp)
{
    PNOTHING_PROP_INTF pNothingFeatures;

    NothingOutputDebugString("NothingSettingsDlgCallback Enter\n");

    switch (uMsg) {

        case PSPCB_CREATE:
            NothingOutputDebugString("NothingSettingsDlgCallback PSPCB_CREATE\n");
            return TRUE;    // return TRUE to continue with creation of page

        case PSPCB_RELEASE:

            //
            // Release the memory that we've previously allocated, because the
            // property page is being deleted.
            //

            NothingOutputDebugString("NothingSettingsDlgCallback PSPCB_RELEASE, Deleting Memory\n");
            pNothingFeatures = (PNOTHING_PROP_INTF) ppsp->lParam;
            if(pNothingFeatures->PIntfDetail) {

                LocalFree(pNothingFeatures->PIntfDetail);

            }
            LocalFree(pNothingFeatures);
            return 0;       // return value ignored

        default:
            break;
    }

    NothingOutputDebugString("NothingSettingsDlgCallback Exit\n");

    return TRUE;

}


///////////////////////////////////////////////////////////////////////////////
//
//  NOTHINGPropPageProvider
//
//      This function gets called by the device manager when it asks for additional
//      property sheet pages. The parameter fAddFunc is the function that we call to
//      add the sheet page to the dialog.
//      This routine gets called because the registry entry "EnumPropPage32" tells
//      the device manager that there is a dll with a exported function that will add
//      a property sheet page.
//
//      Because we want to fail this function (not create the sheet) if the driver
//      doesn't support the private property, we have to do all the work here, that
//      means we open the device and get all the information, then we close the
//      device and return.
//
//  INPUTS:
//
//      pPropPageRequest - points to SP_PROPSHEETPAGE_REQUEST
//      fAddFunc         - function ptr to call to add sheet.
//      lparam           - add sheet functions private data handle.
//
//  OUTPUTS:
//
//      None.
//
//  RETURNS:
//
//      BOOL: FALSE if pages could not be added, TRUE on success.
//
//  NOTES:
//
//
///////////////////////////////////////////////////////////////////////////////
BOOL APIENTRY NOTHINGPropPageProvider(PSP_PROPSHEETPAGE_REQUEST   pPropPageRequest,
	                                  LPFNADDPROPSHEETPAGE	    fAddFunc,
                                      LPARAM					lParam)
{
    PSP_DEVICE_INTERFACE_DETAIL_DATA pDeviceInterfaceDetailData;
    PNOTHING_PROP_INTF               pNothingFeatures;
    PROPSHEETPAGE                    PropSheetPage;
    HPROPSHEETPAGE                   hPropSheetPage;

    NothingOutputDebugString("NOTHINGPropPageProvider Enter\n");

	// Check page requested
    if( pPropPageRequest->PageRequested != SPPSR_ENUM_ADV_DEVICE_PROPERTIES ) {

		return FALSE;
	}

	// Check device info set and data
    if ((!pPropPageRequest->DeviceInfoSet) || (!pPropPageRequest->DeviceInfoData)) {

		return FALSE;
	}

    // Allocate the memory for the Nothing features.
    pNothingFeatures = (PNOTHING_PROP_INTF) LocalAlloc (LPTR, sizeof (NOTHING_PROP_INTF));
    if (!pNothingFeatures) {

        DbgError ("NOTHINGPropPageProvider: LocalAlloc: No Features");
        return FALSE;
    }

    memset(pNothingFeatures,0,sizeof(NOTHING_PROP_INTF));
    
    // Get the device interface detail which return a path to the device
    // driver that we need to open the device.
    if (!GetDeviceInterfaceDetail (pPropPageRequest, &pDeviceInterfaceDetailData)) {

        return FALSE;

    }

    // Get the Nothing features through the private property call.
    if (!GetNothingFeatures (pDeviceInterfaceDetailData, pNothingFeatures)) {

        return FALSE;

    }

    pNothingFeatures->PIntfDetail = pDeviceInterfaceDetailData;

    // We don't need the device interface details any more, get rid of it now!
    // LocalFree (pDeviceInterfaceDetailData);

	// initialize the property page
	PropSheetPage.dwSize		= sizeof(PROPSHEETPAGE);
    PropSheetPage.dwFlags		= PSP_USECALLBACK | PSP_HASHELP;
	PropSheetPage.hInstance		= ghInstance;
	PropSheetPage.pszTemplate	= MAKEINTRESOURCE(DLG_NOTHINGFEATURES);
	PropSheetPage.pfnDlgProc	= NothingDlgProc;
	PropSheetPage.lParam		= (LPARAM) pNothingFeatures;
	PropSheetPage.pfnCallback	= NothingSettingsDlgCallback;

	// create the page and get back a handle
	hPropSheetPage = CreatePropertySheetPage( &PropSheetPage );
    if( !hPropSheetPage ) {

		return FALSE;
	}

	// add the property page
    if( !(*fAddFunc)(hPropSheetPage, lParam)) {

		DestroyPropertySheetPage (hPropSheetPage);
		return FALSE;
	}

    NothingOutputDebugString("NOTHINGPropPageProvider Exit\n");

	return TRUE;
}

